Object subclass: #ObjectWithRoles
	instanceVariableNames: 'roleDictionary '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'RolesV1.0'!


!ObjectWithRoles methodsFor: 'accessing'!

root
   "return the root of the role hierarchy"

   ^self! !

!ObjectWithRoles methodsFor: 'private'!

cancelQualifiedRole: aQualifiedRoleObject
   "remove a qualified role object from the set of roles in the 
    receiver's roleDictionary"

   | aRoleSet aRoleType |

   aRoleType := (aQualifiedRoleObject class) roleTypeName.
   aRoleSet := roleDictionary at: aRoleType.

   aRoleSet remove: aQualifiedRoleObject.

   (aRoleSet isEmpty)
      ifFalse: [roleDictionary at: aRoleType put: aRoleSet]
      ifTrue: [roleDictionary removeKey: aRoleType]!

cancelQualifiedRoleAndSubroles: aQualifiedRoleObject
   "remove a qualified role object and its sub-objects 
    from the receiver's roleDictionary"

   roleDictionary keysDo: 
     [:k | (k inheritsFromRole: ((aQualifiedRoleObject class) roleTypeName))
            ifTrue: [k abandonEntityEquivRolesTo: aQualifiedRoleObject]].
   self cancelQualifiedRole: aQualifiedRoleObject!

cancelRole: aRoleObject
   "remove a role object from the receiver's roleDictionary"

   roleDictionary removeKey: ((aRoleObject class) roleTypeName)!

cancelRoleAndSubroles: aRoleObject
   "remove a role object and its sub-objects from 
    the receiver's roleDictionary"

   roleDictionary keysDo:   
      [:k | (k inheritsFromRole: ((aRoleObject class) roleTypeName))
            ifTrue: [k abandonEntityEquivRolesTo: aRoleObject]].
   self cancelRole: aRoleObject!

initRoleDictionary
   "initialization of instance variable roleDictionary"

   roleDictionary := Dictionary new!

recordNewQualifiedRole: aQualifiedRoleObject
   "insert a new qualified role object into the receiver's roleDictionary"

   | aRoleType aRoleSet |

   aRoleType := (aQualifiedRoleObject class) roleTypeName.

   (self existsAs: (aQualifiedRoleObject class) 
          of: (aQualifiedRoleObject qualifier))
      ifTrue: [self error: 'qualified role already exists']
      ifFalse: [(roleDictionary includesKey: aRoleType)
                    ifFalse: [roleDictionary at: aRoleType put: (Set new)]].

   aRoleSet := roleDictionary at: aRoleType.
   aRoleSet add: aQualifiedRoleObject.
   roleDictionary at: aRoleType put: aRoleSet.!

recordNewRole: aRoleObject
   "insert a new role object into the receiver's roleDictionary"

   (self existsAs: (aRoleObject class))
      ifTrue: [self error: 'role already exists']
      ifFalse: [roleDictionary at: ((aRoleObject class) roleTypeName) 
                                      put: aRoleObject]! !

!ObjectWithRoles methodsFor: 'switching roles'!

as: aRoleType
   "retrieve the receiver's entity-equivalent instance of aRoleType"

   ^roleDictionary at: (aRoleType  roleTypeName)
          ifAbsent: [self error: 'role does not exist']!

as: aQualifiedRoleType of: qualifyingObj
   "retrieve the receiver's entity-equivalent instance of 
    aQualifiedRoleType, qualified by qualifyingObj"

   ^((self as: aQualifiedRoleType) 
         detect: [:o | o qualifier == qualifyingObj]
            ifNone: [self error: 'qualified role does not exist'])! !

!ObjectWithRoles methodsFor: 'testing for role existence'!

existsAs: aRoleType
   "test whether an instance of aRoleType exists that 
    is entity-equivalent to the receiver"

   ^roleDictionary includesKey: (aRoleType roleTypeName)!

existsAs: aQualifiedRoleType of: qualifyingObj
   "test whether an entity-equivalent instance of aQualifiedRoleType, 
    qualified by qualifyingObj, exists"

   ^(roleDictionary includesKey: (aQualifiedRoleType roleTypeName)) and:
        [((roleDictionary at: (aQualifiedRoleType roleTypeName))
             detect: [:o | o qualifier == qualifyingObj] ifNone: []) notNil]! !

!ObjectWithRoles methodsFor: 'testing for entity equivalence'!

entityEquiv: anObject
   "test if the receiver and anObject represent the 
    same real-world entity"

   ^self root == (anObject root)! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ObjectWithRoles class
	instanceVariableNames: ''!


!ObjectWithRoles class methodsFor: 'instance creation'!

new
   "create and initialize a new instance of ObjectWithRoles"

   ^super new initRoleDictionary! !

!ObjectWithRoles class methodsFor: 'role type definition'!

qualifiedRoleSubclass: roleTypeName 
   instanceVariableNames: stringOfInstVarNames
   classVariableNames: stringOfClassVarNames
   poolDictionaries: stringOfPoolNames
   category: categoryNameString
   classOfQualifyingObj: aClass

   "create a new qualified role type of the receiver"

   | newClass |

   (self inheritsFrom: ObjectWithRoles)
      ifFalse: [self error: 'only subclasses of ObjectWithRoles can 
                                 take on roles'].
	
   newClass :=  QualifiedRoleType defQualifiedRoleType: roleTypeName
                         instanceVariableNames: stringOfInstVarNames
                         classVariableNames: stringOfClassVarNames
                         poolDictionaries: stringOfPoolNames
                         category: categoryNameString
                         roleSuperType: self
                         classOfQualifyingObj: aClass.
                                           
   ^newClass!

roleSubclass: roleTypeName 
   instanceVariableNames: stringOfInstVarNames
   classVariableNames: stringOfClassVarNames
   poolDictionaries: stringOfPoolNames
   category: categoryNameString

   "create a new role type of the receiver"

   | newClass |

   (self inheritsFrom: ObjectWithRoles)
      ifFalse: [self error: 'only subclasses of ObjectWithRoles can
                                 take on roles'].
	
   newClass :=  RoleType defRoleType: roleTypeName
                         instanceVariableNames: stringOfInstVarNames
                         classVariableNames: stringOfClassVarNames
                         poolDictionaries: stringOfPoolNames
				 category: categoryNameString
                         roleSuperType: self.
                                           
   ^newClass! !

Object subclass: #RoleType
	instanceVariableNames: 'roleOf root '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'RolesV1.0'!


!RoleType methodsFor: 'accessing'!

roleOf
   "get ancestor"

   ^roleOf!

root
   "get root"

   ^root! !

!RoleType methodsFor: 'private'!

init: ancestor
   "initialize references"

   roleOf := ancestor.
   root := ancestor root.
   root recordNewRole: self!

nullify
   "set all references to nil"

   roleOf := nil.
   root := nil! !

!RoleType methodsFor: 'role abandoning'!

abandon
   "quit this role"

   root cancelRoleAndSubroles: self! !

!RoleType methodsFor: 'switching roles'!

as: aRoleType
   "retrieve entity-equivalent instance of aRoleType"

   ^root as: aRoleType!

as: aQualifiedRoleType of: qualifyingObj
   "retrieve entity-equivalent instance, qualified by 
    qualifyingObj, of aQualifiedRoleType"

   ^root as: aQualifiedRoleType of: qualifyingObj! !

!RoleType methodsFor: 'testing for entity equivalence'!

entityEquiv: anObject
   "test if the receiver and anObject represent 
    the same real-world entity"

   ^root == (anObject root)! !

!RoleType methodsFor: 'testing for role existence'!

existsAs: aRoleType
   "check whether an entity-equivalent instance
    of aRoleType exists"

   ^root existsAs: aRoleType!

existsAs: aQualifiedRoleType of: qualifyingObj
   "check whether an entity-equivalent instance of 
    aQualifiedRoleType, qualified by qualifyingObj exists"

   ^root existsAs: aQualifiedRoleType of: qualifyingObj! !

!RoleType methodsFor: 'inheritance'!

doesNotUnderstand: aMessage
   "delegate every message not understood to the ancestor"

   ^roleOf perform: (aMessage selector) 
              withArguments: (aMessage arguments)! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

RoleType class
	instanceVariableNames: ''!


!RoleType class methodsFor: 'instance creation'!

newRoleOf: anObject
   "create a new role for anObject"

   (anObject isKindOf: (self roleSuperType))
      ifFalse: [self error: 'super object of wrong type'].
   ^self new init: anObject! !

!RoleType class methodsFor: 'role type definition'!

defRoleType: roleTypeName 
   instanceVariableNames: stringOfInstVarNames
   classVariableNames: stringOfClassVarNames
   poolDictionaries: stringOfPoolNames
   category: categoryNameString
   roleSuperType: nameOfRoleSupertype

   "create a new role-subclass of the class RoleType"

   | newClass nameOfRoleSuperType |

   nameOfRoleSuperType := Smalltalk at: nameOfRoleSupertype.

   ((nameOfRoleSuperType inheritsFrom: QualifiedRoleType) and:
      [self ~= QualifiedRoleType])
         ifTrue: [self error: 'you need to subclass QualifiedRoleType'].

   (self ~= QualifiedRoleType)
      ifTrue: [(((nameOfRoleSuperType superclass) = RoleType) |
         (nameOfRoleSuperType inheritsFrom: ObjectWithRoles))
            ifFalse: [self error: 'RoleSuperType must be a 
                                      subclass of ObjectWithRoles or of RoleType']].
	
   newClass := super  subclass: roleTypeName
                        instanceVariableNames: stringOfInstVarNames
                        classVariableNames: stringOfClassVarNames
                        poolDictionaries: stringOfPoolNames
                        category: categoryNameString.

   newClass addClassVarName: 'RoleSuperType'.
   newClass addClassVarName: 'RoleTypeName'.
   newClass class compile: ('initRoleSuperType: param 
      RoleSuperType := param') classified: 'class initialization'.
   newClass class compile: ('initRoleTypeName: param 
      RoleTypeName := param') classified: 'class initialization'.
   newClass initRoleSuperType: nameOfRoleSuperType.
   newClass initRoleTypeName: newClass. 
   newClass class removeCategory: 'class initialization'.
   newClass class compile: ('roleSuperType ^RoleSuperType') 
      classified: 'accessing class variables'.
   newClass class compile: ('roleTypeName ^RoleTypeName') 
      classified: 'accessing class variables'.
   ^newClass!

inheritsFromRole: aRoleType
   "answer whether aRoleType is on the receiver's 
    role-hierarchy chain"
	
   ^((self roleSuperType = aRoleType) |
      ((self roleSuperType inheritsFrom: RoleType) and:
         [self roleSuperType inheritsFromRole: aRoleType]))!

qualifiedRoleSubclass: roleTypeName
   instanceVariableNames: stringOfInstVarNames
   classVariableNames: stringOfClassVarNames
   poolDictionaries: stringOfPoolNames
   category: categoryNameString
   classOfQualifyingObj: aClass
 
   "create a new qualified role type of the receiver"
 
   | newClass |
 
   newClass :=  QualifiedRoleType defQualifiedRoleType: roleTypeName
                          instanceVariableNames: stringOfInstVarNames
                          classVariableNames: stringOfClassVarNames
                          poolDictionaries: stringOfPoolNames
                          category: categoryNameString
                          roleSuperType: self
                          classOfQualifyingObj: aClass.

    ^newClass!

roleSubclass: roleTypeName 
   instanceVariableNames: stringOfInstVarNames
   classVariableNames: stringOfClassVarNames
   poolDictionaries: stringOfPoolNames
   category: categoryNameString

   "create a new role type of the receiver"

   | newClass |

   newClass :=  RoleType defRoleType: roleTypeName
                         instanceVariableNames: stringOfInstVarNames
                         classVariableNames: stringOfClassVarNames
                         poolDictionaries: stringOfPoolNames
                         category: categoryNameString
                         roleSuperType: self.
                                           
   ^newClass! !

!RoleType class methodsFor: 'private'!

abandonEntityEquivRolesTo: anObject
   "remove all occurences of roleType that are 
    entity-equivalent to anObject"

   (anObject root) cancelRole: (anObject as: self)! !

RoleType subclass: #QualifiedRoleType
	instanceVariableNames: 'qualifier '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'RolesV1.0'!


!QualifiedRoleType methodsFor: 'accessing'!

qualifier
   "return the value of the instance variable qualifier"

   ^qualifier! !

!QualifiedRoleType methodsFor: 'private'!

init: ancestor 
   "not applicable to instances of QualifiedRoleType"

   self shouldNotImplement!

init: ancestor qualifiedBy: qualifyingObj
   "initialize the receiver's references in the role hierarchy"

   roleOf := ancestor.
   root := ancestor root.
   qualifier := qualifyingObj.
   root recordNewQualifiedRole: self! !

!QualifiedRoleType methodsFor: 'role abandoning'!

abandon
   "quit this qualified role"

   root cancelQualifiedRoleAndSubroles: self! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

QualifiedRoleType class
	instanceVariableNames: ''!


!QualifiedRoleType class methodsFor: 'instance creation'!

newRoleOf: anObject 
   "not applicable to instances of QualifiedRoleType"
	
   self shouldNotImplement!

newRoleOf: anObject qualifiedBy: qualifyingObj
   "create an instance of the receiver qualified by 
    qualifyingObj as a new role of anObject"

   (anObject isKindOf: (self roleSuperType))
      ifFalse: [self error: 'super object of wrong type'].
   (qualifyingObj isKindOf: (self classOfQualifyingObj))
      ifFalse: [self error: 'qualifying object of wrong type'].
   ^self new init: anObject qualifiedBy: qualifyingObj! !

!QualifiedRoleType class methodsFor: 'role type definition'!

defQualifiedRoleType: roleTypeName 
   instanceVariableNames: stringOfInstVarNames
   classVariableNames: stringOfClassVarNames
   poolDictionaries: stringOfPoolNames
   category: categoryNameString
   roleSuperType: nameOfRoleSuperType
   classOfQualifyingObj: aClass

   "create a new subclass of the class QualifiedRoleType"

   | newClass |

   self = QualifiedRoleType
      ifFalse: [self shouldNotImplement].
   ((nameOfRoleSuperType superclass) = QualifiedRoleType) |
      ((nameOfRoleSuperType superclass) = RoleType) |
         (nameOfRoleSuperType inheritsFrom: ObjectWithRoles)
           ifFalse: [self error: 'RoleSuperType must be a subclass of
                    ObjectWithRoles, of RoleType, or of QualifiedRoleType'].
	
   newClass := super  defRoleType: roleTypeName 
                        instanceVariableNames: stringOfInstVarNames
                        classVariableNames: stringOfClassVarNames
                        poolDictionaries: stringOfPoolNames
                        category: categoryNameString
                        roleSuperType: nameOfRoleSuperType.

   newClass addClassVarName: 'ClassOfQualifyingObj'.
   newClass class compile: ('initClassOfQualifyingObj: param 
      ClassOfQualifyingObj := param') classified: 'class initialization'.
   newClass initClassOfQualifyingObj: (Smalltalk at: aClass).
   newClass class removeCategory: 'class initialization'.
   newClass class compile: ('classOfQualifyingObj ^ClassOfQualifyingObj') 
      classified: 'accessing class variables'.
   ^ newClass!

defRoleType: roleTypeName 
   instanceVariableNames: stringOfInstVarNames
   classVariableNames: stringOfClassVarNames
   poolDictionaries: stringOfPoolNames
   category: categoryNameString
   roleSuperType: nameOfRoleSuperType

   "not applicable to qualified role types"

   self shouldNotImplement!

qualifiedRoleSubclass: roleTypeName 
   instanceVariableNames: stringOfInstVarNames
   classVariableNames: stringOfClassVarNames
   poolDictionaries: stringOfPoolNames
   category: categoryNameString
   classOfQualifyingObj: aClass

   "create a new role type of the receiver"

   | newClass |

   newClass :=  QualifiedRoleType defQualifiedRoleType: roleTypeName
                         instanceVariableNames: stringOfInstVarNames
                         classVariableNames: stringOfClassVarNames
                         poolDictionaries: stringOfPoolNames
                         category: categoryNameString
                         roleSuperType: self
                         classOfQualifyingObj: aClass.
                                           
   ^newClass!

roleSubclass: roleTypeName 
   instanceVariableNames: stringOfInstVarNames
   classVariableNames: stringOfClassVarNames
   poolDictionaries: stringOfPoolNames
   category: categoryNameString

   "not applicable to qualified role types"

   self shouldNotImplement! !

!QualifiedRoleType class methodsFor: 'private'!

abandonEntityEquivRolesTo: anObject
   "remove all occurences of QualifiedRoleType that
    are entity-equivalent to anObject"

   (anObject as: self)
      do: [:o | (anObject root) cancelQualifiedRole: o]! !
